import argparse
import pathlib
from typing import Union, List
from typing import Optional as OptionalType, NamedTuple, Callable
import dataclasses
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from voluptuous import Schema, Optional, And, Coerce, Invalid, Range

from tclint.violations import Rule


class ExcludePattern(NamedTuple):
    """Exclude patterns are applied relative to a certain root. This dataclass is used
    to bundle the two."""

    pattern: str
    root: pathlib.Path


@dataclasses.dataclass
class Config:
    """This dataclass defines the supported Config fields and their default
    values. It provides an external interface for accessing config values.

    The type annotations defined here are fairly loose - more specific type
    validation (and normalization) is defined by `validators` below.
    """

    # TODO: switch to `exclude: List[ExcludePattern]`
    exclude: List[str] = dataclasses.field(default_factory=list)
    ignore: List[Rule] = dataclasses.field(default_factory=list)
    commands: OptionalType[pathlib.Path] = dataclasses.field(default=None)
    extensions: List[str] = dataclasses.field(
        default_factory=lambda: ["tcl", "sdc", "xdc", "upf"]
    )
    style_indent: OptionalType[Union[str, int]] = dataclasses.field(default=None)
    style_line_length: int = dataclasses.field(default=100)
    style_max_blank_lines: int = dataclasses.field(default=2)
    style_indent_namespace_eval: bool = dataclasses.field(default=True)
    style_spaces_in_braces: bool = dataclasses.field(default=False)

    def apply_cli_args(self, args):
        args_dict = vars(args)
        for field in dataclasses.fields(self):
            if field.name in args_dict and args_dict[field.name] is not None:
                setattr(self, field.name, args_dict[field.name])

        # Special arguments that aren't handled automatically
        if "extend_exclude" in args_dict and args_dict["extend_exclude"] is not None:
            self.exclude.extend(args_dict["extend_exclude"])

        if "extend_ignore" in args_dict and args_dict["extend_ignore"] is not None:
            self.ignore.extend(args_dict["extend_ignore"])

    def get_indent(self) -> str:
        """Get indent setting as string.

        This helper does two things. One, it's a helpful utility to factor out the logic
        required for calculating the indent. Two, it lets us ergonomically store if the
        indentation is not set in style_indent, which the LSP relies on.
        """
        if self.style_indent is None:
            # Default indent
            return " " * 4
        elif self.style_indent == "tab":
            return "\t"
        elif isinstance(self.style_indent, int):
            return " " * self.style_indent
        elif isinstance(self.style_indent, tuple):
            return " " * self.style_indent[0]

        # Should be unreachable, validated on ingestion of config
        raise ValueError(
            f"unexpected value for config.style_indent: {self.style_indent}"
        )

    def get_indent_mixed_tab_size(self) -> int:
        if isinstance(self.style_indent, tuple):
            return self.style_indent[1]
        return 0


# Validators using `voluptuous` library that check and normalize config inputs.
# Used for checking both config files as well as config-related CLI args.

# Using these for CLI args adds a constraint that all non-boolean validators
# need to be able to normalize a value from a string. This means one could put
# e.g. a string representation of a list into a .toml config file, but we shouldn't
# document this, since it won't be considered stable behavior.


def _str2list(s):
    """Handles string-to-list normalization."""
    if isinstance(s, str):
        if s == "":
            return []
        return [v.strip() for v in s.split(",")]
    return s


def parse_mixed(v: str) -> tuple[int, int]:
    """Parse --indent=mixed,<s>,<t>."""
    s = v.split(",")
    if not (len(s) == 3 and s[0] == "mixed" and s[1].isdigit() and s[2].isdigit()):
        raise ValueError()
    return (int(s[1]), int(s[2]))


# Define validators as module variables so they can be reused for config file schema
# validation and CLI argument parsing.

# note: it's ok if paths don't exist - allows for generic
# configurations with directories like .git/ excluded
_validate_exclude = _str2list
_validate_ignore = And(
    _str2list,
    [
        Coerce(Rule, msg="invalid rule ID"),
    ],
)
_validate_commands = Coerce(pathlib.Path)
_validate_extensions = _str2list
_validate_style_indent = Coerce(
    lambda v: (
        v
        if v == "tab"
        else (
            int(v)
            if isinstance(v, int) or (isinstance(v, str) and v.isdigit())
            else parse_mixed(v)
        )
    ),
    msg="expected integer, 'tab', or 'mixed',integer,integer",
)
_validate_style_line_length = Coerce(int)
_validate_style_max_blank_lines = And(
    Coerce(int),
    Range(min=1),
)
_validate_style_indent_namespace_eval = bool
_validate_style_spaces_in_braces = bool


def _error_if_dynamic_plugin(p: pathlib.Path):
    if p.suffix == ".py":
        raise Invalid(
            "dynamic plugins cannot be specified via config file for security"
            " reasons. Use --commands instead."
        )
    return p


def _validate_config(config):
    """Validates dictionary read from TOML config file. Individual value validators
    are implemented in the module-level variables above; this defines the actual
    structure of the schema."""
    base_config = {
        Optional("ignore"): _validate_ignore,
        # tclint rejects paths to dynamic plugins in the config file. This restriction
        # is designed to make it explicit when tclint is executing external code.
        Optional("commands"): And(_validate_commands, _error_if_dynamic_plugin),
        Optional("style"): {
            Optional("indent"): _validate_style_indent,
            Optional("line-length"): _validate_style_line_length,
            Optional("max-blank-lines"): _validate_style_max_blank_lines,
            Optional("indent-namespace-eval"): _validate_style_indent_namespace_eval,
            Optional("spaces-in-braces"): _validate_style_spaces_in_braces,
        },
    }

    schema = Schema({
        # exclude and extensions can only be used in global context
        Optional("exclude"): _validate_exclude,
        Optional("extensions"): _validate_extensions,
        **base_config,
        Optional("fileset"): Schema(
            [{"paths": [Coerce(pathlib.Path)], **base_config}], required=True
        ),
    })

    try:
        return schema(config)
    except Invalid as e:
        if not e.path:
            raise ConfigError(e.error_message)

        # Stringify error path to my own taste.
        path = []
        for item in e.path:
            if isinstance(item, int):
                # Brackets around indices
                if len(path) > 0:
                    path[-1] += f"[{item}]"
                else:
                    path.append(f"[{item}]")
            else:
                path.append(str(item))

        raise ConfigError(f"{e.error_message} ({'.'.join(path)})")


def _argparsify(validator: Callable) -> Callable:
    """Wrapper that applies a voluptuous-style validator and is compatible with
    argparse's type argument."""

    def func(s):
        try:
            return Schema(validator)(s)
        except Invalid as e:
            raise argparse.ArgumentTypeError(str(e))

    return func


def _add_bool(group, parser, dest, yes_flag, no_flag):
    mutex_group = group.add_mutually_exclusive_group(required=False)
    mutex_group.add_argument(yes_flag, dest=dest, action="store_true")
    mutex_group.add_argument(no_flag, dest=dest, action="store_false")
    parser.set_defaults(**{dest: None})


def setup_common_config_cli_args(config_group):
    config_group.add_argument(
        "--exclude", type=_argparsify(_validate_exclude), metavar='"path1, path2, ..."'
    )
    config_group.add_argument(
        "--extend-exclude",
        type=_argparsify(_validate_exclude),
        metavar='"path1, path2, ..."',
    )
    config_group.add_argument(
        "--extensions",
        type=_argparsify(_validate_extensions),
        metavar='"tcl, xdc, ..."',
    )
    config_group.add_argument(
        "--commands", type=_argparsify(_validate_commands), metavar="<path>"
    )


def setup_config_cli_args(parser):
    """This method defines config-related CLI arguments.

    The destvars of these switches should match the fields of Config.
    """
    config_group = parser.add_argument_group("configuration arguments")

    config_group.add_argument(
        "--ignore", type=_argparsify(_validate_ignore), metavar='"rule1, rule2, ..."'
    )
    config_group.add_argument(
        "--extend-ignore",
        type=_argparsify(_validate_ignore),
        metavar='"rule1, rule2, ..."',
    )
    setup_common_config_cli_args(config_group)
    config_group.add_argument(
        "--style-line-length",
        type=_argparsify(_validate_style_line_length),
        metavar="<line_length>",
    )


def setup_tclfmt_config_cli_args(parser):
    """This method defines the subset of config-related CLI arguments used by tclfmt.

    The destvars of these switches should match the fields of Config.
    """
    config_group = parser.add_argument_group("configuration arguments")

    setup_common_config_cli_args(config_group)

    config_group.add_argument(
        "--indent",
        type=_argparsify(_validate_style_indent),
        metavar="<indent>",
        dest="style_indent",
    )
    config_group.add_argument(
        "--max-blank-lines",
        type=_argparsify(_validate_style_max_blank_lines),
        metavar="<max_blank_lines>",
        dest="style_max_blank_lines",
    )
    _add_bool(
        config_group,
        parser,
        "style_indent_namespace_eval",
        "--indent-namespace-eval",
        "--no-indent-namespace-eval",
    )
    _add_bool(
        config_group,
        parser,
        "style_spaces_in_braces",
        "--spaces-in-braces",
        "--no-spaces-in-braces",
    )


def _flatten(d, prefix=None):
    """Flattens TOML config dictionary structure to match the flat set of fields
    expected by Config dataclass."""
    if prefix is None:
        prefix = []

    flat = {}
    for k, v in d.items():
        if isinstance(v, dict):
            flat.update(_flatten(v, prefix=prefix + [k]))
        else:
            flat["_".join(prefix + [k]).replace("-", "_")] = v

    return flat


class RunConfig:
    """Class that holds information about both global and fileset configs. User
    code can get a Config object that applies to a particular file by calling
    get_from_path() and supplying that file's path."""

    def __init__(self, global_config=None, fileset_configs=None):
        if global_config is not None:
            self._global_config = global_config
        else:
            self._global_config = Config()

        self._fileset_configs = [
            # ([pathlib.Path...], Config])
        ]
        if fileset_configs is not None:
            self._fileset_configs = fileset_configs

    @property
    def exclude(self):
        return self._global_config.exclude

    @property
    def extensions(self):
        return self._global_config.extensions

    @classmethod
    def from_dict(cls, config_dict: dict, root: pathlib.Path):
        config_dict = _validate_config(config_dict)
        try:
            fileset_config_dicts = config_dict.pop("fileset")
        except KeyError:
            fileset_config_dicts = []

        config_dict = _flatten(config_dict)
        global_config = Config(**config_dict)

        fileset_configs = []
        for fileset_config in fileset_config_dicts:
            paths = []
            for path in fileset_config.pop("paths"):
                if not path.is_absolute():
                    path = root / path
                paths.append(path.resolve())

            fileset_config = _flatten(fileset_config)

            # pull in default values from global config
            full_fileset_config = config_dict.copy()
            full_fileset_config.update(fileset_config)

            fileset_configs.append((paths, Config(**full_fileset_config)))

        return cls(global_config, fileset_configs)

    @classmethod
    def from_path(cls, path: Union[str, pathlib.Path], root: pathlib.Path):
        path = pathlib.Path(path)

        if not path.exists():
            raise FileNotFoundError

        with open(path, "rb") as f:
            try:
                data = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                raise ConfigError(f"{path}: {e}")

        try:
            return cls.from_dict(data, root)
        except ConfigError as e:
            raise ConfigError(f"{path}: {e}")

    @classmethod
    def from_pyproject(cls, directory=None):
        if directory is None:
            directory = pathlib.Path(".")
        else:
            directory = pathlib.Path(directory)

        path = directory / "pyproject.toml"

        if not path.exists():
            raise FileNotFoundError

        with open(path, "rb") as f:
            data = tomllib.load(f)

        tclint_config = data.get("tool", {})["tclint"]

        try:
            return cls.from_dict(tclint_config, directory)
        except ConfigError as e:
            raise ConfigError(f"pyproject.toml: {e}")

    def get_for_path(self, path) -> Config:
        if path is None:
            return self._global_config

        path = path.resolve()
        for fileset_paths, config in self._fileset_configs:
            for fileset_path in fileset_paths:
                if path.is_relative_to(fileset_path):
                    return config

        return self._global_config

    def apply_cli_args(self, args):
        self._global_config.apply_cli_args(args)
        for _, fileset_config in self._fileset_configs:
            fileset_config.apply_cli_args(args)


class ConfigError(Exception):
    pass


DEFAULT_CONFIGS = ("tclint.toml", ".tclint")


def get_config(
    config_path: OptionalType[pathlib.Path], root: pathlib.Path
) -> OptionalType[RunConfig]:
    """Loads a config file.

    If `config_path` is supplied, attempts to read config file from this path. If the
    path can't be found, raises a ConfigError.

    Otherwise, attempts to read config from `root`/{tclint.toml, .tclint,
    pyproject.toml} (in that order). If none of these files can be found, returns None.

    `root` is also used to resolve some relative paths in the config file.
    """
    # user-supplied
    if config_path is not None:
        try:
            return RunConfig.from_path(config_path, root)
        except FileNotFoundError:
            raise ConfigError(f"path {config_path} doesn't exist")

    for path in DEFAULT_CONFIGS:
        try:
            return RunConfig.from_path(root / path, root)
        except FileNotFoundError:
            pass

    try:
        return RunConfig.from_pyproject(directory=root)
    except ConfigError as e:
        raise e
    except (FileNotFoundError, tomllib.TOMLDecodeError, KeyError):
        # just skip if file doesn't exist, contains TOML errors, or tclint key not found
        pass

    return None



import pytest

from tclint.parser import Parser
from tclint.syntax_tree import Script, Command, BareWord, QuotedWord


@pytest.mark.parametrize(
    "command,valid",
    [
        ("namespace path", True),
        (r"namespace path {path1 path2}", True),
        ("namespace path path1 path2", False),
        ("namespace upvar", False),
        ("namespace upvar my_namespace", True),
        ("namespace upvar my_namespace var0 var1 var2", True),
        ("namespace unknown", True),
        ("namespace unknown puts hello", False),
        ("break", True),
        ("break asdf", False),
        ("foreach", False),
        (r"foreach $iters {}", False),
        (r"foreach {*}$iters {}", True),
    ],
)
def test_validation(command, valid):
    parser = Parser()
    parser.parse(command)

    assert len(parser.violations) == 0 if valid else 1


def test_foreach_arg_expansion_only():
    parser = Parser()
    parser.parse("foreach {*}$args")

    assert len(parser.violations) == 1
    assert not str(parser.violations[0]).startswith("1:1: insufficient args")
    assert str(parser.violations[0]).startswith("1:1: expected braced word")


def test_namespace_unknown_script():
    script = r"""namespace unknown {
        puts "hello"
    }"""
    parser = Parser()
    tree = parser.parse(script)
    assert tree == Script(
        Command(
            BareWord("namespace"),
            BareWord("unknown"),
            Script(Command(BareWord("puts"), QuotedWord(BareWord("hello")))),
        )
    )


def test_if_valid():
    parser = Parser()

    for script in [
        "if 1 {}",
        "if 1 {} {}",
        "if 1 {} elseif 1 {}",
        "if 1 {} elseif 1 {} {}",
        "if 1 then {}",
        "if 1 {} else {}",
        "if 1 {} elseif 1 then {}",
        "if 1 {} elseif 1 {} else {}",
    ]:
        parser.parse(script)
        assert len(parser.violations) == 0


def test_if_invalid():
    parser = Parser()

    n = 0
    for script in [
        "if 1",
        "if 1 {} else",
        "if 1 {} else {} {}",
        "if 1 {} elseif",
        "if 1 {} elseif 1",
        "if 1 {} elseif 1 {} else",
        "if 1 {} elseif 1 {} else {} {}",
    ]:
        n += 1
        parser.parse(script)
        assert len(parser.violations) == n



from pathlib import Path

from tclint.symbol_table import SymbolTable, SymbolTableBuilder
import tclint.syntax_tree as ast
from tclint.parser import Parser

MY_DIR = Path(__file__).parent.resolve()


def test_symtab_builder():
    source_text = Path(MY_DIR / "data" / "symbols.tcl").read_text()
    tree = Parser().parse(source_text)
    symtab = SymbolTableBuilder().build(tree)

    hello_world = symtab.lookup_proc_definitions("hello_world")
    assert len(hello_world) == 1
    assert hello_world[0].pos == (2, 6)


def test_static_proc_name():
    symtab = SymbolTable()
    command = ast.Command(
        ast.BareWord("proc"), ast.BareWord("foo"), ast.List(), ast.Script()
    )
    symtab.add_proc_definition(command)

    foo = symtab.lookup_proc_definitions("foo")
    assert len(foo) == 1
    assert foo[0].value == "foo"


def test_dynamic_proc_name():
    symtab = SymbolTable()
    # Result of parsing `proc $name {} {}`
    command = ast.Command(
        ast.BareWord("proc"), ast.VarSub("name"), ast.List(), ast.Script()
    )
    symtab.add_proc_definition(command)

    assert symtab.lookup_proc_definitions("name") == []
    assert symtab.lookup_proc_definitions("$name") == []



"""Helpers for checking command arguments."""

from __future__ import annotations
from collections.abc import Callable
from typing import List, Optional, Union, Tuple, TYPE_CHECKING

from tclint.syntax_tree import ArgExpansion, QuotedWord, BracedWord, BareWord, Node

# This lets us use Parser in type annotations without introducing a cyclic dependency.
if TYPE_CHECKING:
    from tclint.parser import Parser


class CommandArgError(Exception):
    pass


def arg_count(args: List[Node], parser: Parser) -> Tuple[int, bool]:
    """Returns the number of arguments in args, taking {*} into account.

    If an argument list contains an argument expansion operator that cannot be
    statically expanded, the second return value is True, and the count is the minimum
    possible number of arguments. Otherwise, the return value is False and the count
    reflects the exact number of arguments.

    This function should always be used for validating argument count, rather than
    relying on `len(args)`.
    """

    # TODO: Replace this with or add a similar `expand_args` function that returns an
    # expanded argument list. One tricky thing is we want to handle cases like:
    # - foreach {*}$iters { ...body... }
    # - catch {puts "my script"} {*}$catchopts
    # With something like a list structure we can either forward or reverse index to
    # still parse the body even with an unexpanded {*}.

    # TODO: Add a violation that flags `{*}{a b c}` for rewriting as `a b c`. A future
    # version of tclfmt that allows rewrites that break the syntax tree could do this
    # automatically.

    arg_count = 0
    has_arg_expansion = False
    for arg in args:
        if isinstance(arg, ArgExpansion):
            if arg.contents is None:
                has_arg_expansion = True
                continue
            arg_count += len(parser.parse_list(arg).children)
        else:
            arg_count += 1

    return arg_count, has_arg_expansion


def check_count(command, min=None, max=None, args_name="args"):
    def check(args, parser):
        if min is None and max is None:
            return None

        count, has_arg_expansion = arg_count(args, parser)

        if not has_arg_expansion and min == max and count != min:
            raise CommandArgError(
                f"wrong # of {args_name} for {command}: got {count}, expected {min}"
            )

        if not has_arg_expansion and min is not None and count < min:
            raise CommandArgError(
                f"not enough {args_name} for {command}: got {count}, expected at least"
                f" {min}"
            )

        if max is not None and count > max:
            raise CommandArgError(
                f"too many {args_name} for {command}: got {count}, expected no more"
                f" than {max}"
            )

        return None

    return check


def eval(args: List[Node], parser: Parser, command: str) -> List[Node]:
    if len(args) > 1 and any(isinstance(arg, (QuotedWord, BracedWord)) for arg in args):
        # Slightly odd restriction, but our syntax tree doesn't have a great way
        # to handle this case. We require each command argument to correspond to
        # one child node, but multiple quoted or braced word arguments can be
        # combined into a single subcommand when interpreted eval-style. This
        # requirement exists to facilitate style checking, if we had a separate
        # CST for style checks and AST for logical checks we may be able to
        # handle it.

        raise CommandArgError(
            f"unable to parse multiple {command} arguments when one includes a braced"
            " or quoted word"
        )

    # Construct the body of the eval taking whitespace into account to ensure we get
    # style checking.

    eval_script = ""
    prev_arg_end_pos = None
    for arg in args:
        contents = arg.contents
        if contents is None:
            # TODO: flag sort of eval-specific violation? Common patterns will
            # often trigger this, and it seems useful to be able to turn it off
            raise CommandArgError(
                f"{command} received an argument with a substitution, unable to parse"
                " its arguments"
            )

        if prev_arg_end_pos is not None:
            if prev_arg_end_pos[0] != arg.line:
                # If we have multiple args on the same line, we know there must be a
                # backslash newline. Add it so the parsing works.
                eval_script += "\\\n" * (arg.line - prev_arg_end_pos[0])
                eval_script += " " * (arg.col - 1)
            else:
                eval_script += " " * (arg.col - prev_arg_end_pos[1])
        eval_script += contents

        prev_arg_end_pos = arg.end_pos

    script = parser.parse(eval_script, pos=(args[0].pos))
    script.end_pos = args[-1].end_pos

    return [script]


def check_command(
    command: str,
    args: List[Node],
    parser: Parser,
    command_spec: Union[Callable, dict, None],
) -> Optional[List[Node]]:
    if command_spec is None:
        return None

    if isinstance(command_spec, dict):
        return check_arg_spec(command, args, parser, command_spec)

    return command_spec(args, parser)


def check_arg_spec(
    command: str, args: List[Node], parser: Parser, arg_spec: dict
) -> Optional[List[Node]]:
    if "subcommands" in arg_spec:
        subcommands = arg_spec["subcommands"]
        try:
            subcommand = args[0].contents
        except IndexError:
            subcommand = None

        if subcommand in subcommands:
            new_args = check_command(
                f"{command} {subcommand}", args[1:], parser, subcommands[subcommand]
            )
            if new_args is None:
                return new_args
            return args[0:1] + new_args

        if "" in subcommands:
            return check_command(command, args, parser, subcommands[""])

        if subcommand is not None:
            msg = f"invalid subcommand for {command}: got {subcommand}"
        else:
            msg = f"no subcommand provided for {command}"

        raise CommandArgError(f"{msg}, expected one of {', '.join(subcommands.keys())}")

    switches = arg_spec["switches"]
    args_allowed = set(switches)
    args_required = {switch for switch in switches if switches[switch]["required"]}
    positional_args = []

    args = list(args)
    if not switches:
        positional_args = args
        args = []

    while len(args) > 0:
        arg = args.pop(0)

        # To facilitate better error messages, we expect that switches are always
        # specified as BareWords that start with "-" or ">". This lets us throw an
        # error when a switch-like thing doesn't match any supported arguments,
        # rather than counting it towards the positional arguments (which usually
        # ends up in a vague "too many arguments" error). To make tclint interpret a
        # switch-like word as a positional argument, users should wrap it in "", and
        # any switches should be BareWords.
        contents = arg.contents
        if not (isinstance(arg, BareWord) and contents and contents[0] in {"-", ">"}):
            positional_args.append(arg)
            continue

        # TODO check required arguments
        if contents in args_allowed:
            if switches[contents]["value"]:
                try:
                    args.pop(0)
                except IndexError:
                    raise CommandArgError(
                        f"invalid arguments for {command}: expected value after"
                        f" {contents}"
                    )
            if not switches[contents]["repeated"]:
                args_allowed.remove(contents)
            if contents in args_required:
                args_required.remove(contents)
        elif contents in arg_spec:
            raise CommandArgError(f"duplicate argument for {command}: {contents}")
        else:
            prefix_matches = []
            for switch in switches:
                if switch.startswith(contents):
                    prefix_matches.append(switch)

            if len(prefix_matches) == 1:
                raise CommandArgError(
                    f"shortened argument for {command}: expand {contents} to"
                    f" {prefix_matches[0]}"
                )

            if len(prefix_matches) > 1:
                raise CommandArgError(
                    f"ambiguous argument for {command}: {contents} could be any of"
                    f" {', '.join(prefix_matches)}"
                )

            raise CommandArgError(f"unrecognized argument for {command}: {contents}")

    if len(args_required) > 1:
        raise CommandArgError(
            f"missing required arguments for {command}: {', '.join(args_required)}"
        )
    elif len(args_required) == 1:
        raise CommandArgError(
            f"missing required argument for {command}: {args_required.pop()}"
        )

    min_positionals = 0
    max_positionals: Optional[int] = 0
    required_positionals = []

    for positional in arg_spec["positionals"]:
        if positional["value"]["type"] == "variadic":
            max_positionals = None

        if positional["required"]:
            min_positionals += 1
            required_positionals.append(positional["name"])
        if max_positionals is not None:
            max_positionals += 1

    # Check for arg expansions first
    count, has_arg_expansion = arg_count(positional_args, parser)

    # Only check for missing arguments if there are no arg expansions
    if not has_arg_expansion and count < min_positionals:
        missing_positionals = required_positionals[count:]
        if len(missing_positionals) == 1:
            raise CommandArgError(
                f"missing required positional argument for {command}: "
                f"{missing_positionals[0]}"
            )
        else:
            raise CommandArgError(
                f"missing required positional arguments for {command}: "
                f"{', '.join(missing_positionals)}"
            )

    # Always check for too many arguments
    if max_positionals is not None and count > max_positionals:
        raise CommandArgError(
            f"too many positional args for {command}: "
            f"got {count}, expected no more than {max_positionals}"
        )

    # No need for check_count here as we've already done the checks

    return None



import logging
from collections import defaultdict
from typing import List, DefaultDict, Union

from tclint.syntax_tree import Visitor, Command, CommandSub, Node, Script


class SymbolTable:
    """Holds a symbol table (links symbols to nodes)."""

    def __init__(self):
        self.proc_def: DefaultDict[str, list[Node]] = defaultdict(list)

    def add_proc_definition(self, command: Command) -> None:
        """Add definition of procedure"""
        # command holds the "proc" keyword, so the proc name is 1st argument
        proc_name_node = command.args[0]
        proc_name = proc_name_node.contents
        if not proc_name:
            return
        logging.debug(
            f"Definition of proc '{proc_name}' at {proc_name_node._pos_str()}"
        )
        self.proc_def[proc_name].append(proc_name_node)

    def lookup_proc_definitions(self, symbol_text: str) -> List[Node]:
        """Lookup definitions of the procedure pointed at by node"""
        if symbol_text is None or symbol_text not in self.proc_def:
            return []
        return self.proc_def[symbol_text]


class SymbolTableBuilder(Visitor):
    """Builds a symbol table."""

    def __init__(self):
        self.table = SymbolTable()

    def build(self, tree: Union[CommandSub, Script]) -> SymbolTable:
        """Run the builder visitor through the syntax tree, building a table."""
        tree.accept(self, recurse=True)
        return self.table

    def visit_command(self, command: Command) -> None:
        if command.routine.contents == "proc":
            self.table.add_proc_definition(command)



import codecs
from collections import defaultdict
import os
import pathlib
import re
from typing import Callable, List, Optional

import pathspec

from tclint.config import ExcludePattern


def register_codec_warning(name):
    def replace_with_warning_handler(e):
        # TODO: formal warning mechanism, include path
        print("Warning: non-unicode characters in file, replacing with �")
        return codecs.replace_errors(e)

    codecs.register_error(name, replace_with_warning_handler)


def make_exclude_filter(
    exclude_patterns: List[ExcludePattern],
) -> Callable[[pathlib.Path], bool]:
    # Transform patterns into a data structure keyed on root.
    patterns_by_root = defaultdict(list)
    for pattern, root in exclude_patterns:
        # I think this is escaping #, which would otherwise be treated like a comment.
        # Not 100% sure though, I originally wrote this a while ago.
        pattern = re.sub(r"^\s*#", r"\#", pattern)
        patterns_by_root[root.resolve()].append(pattern)

    compiled_patterns = {}
    for root in patterns_by_root.keys():
        patterns = patterns_by_root[root]
        spec = pathspec.PathSpec.from_lines("gitwildmatch", patterns)
        compiled_patterns[root] = spec

    def is_excluded(path: pathlib.Path) -> bool:
        abspath = path.resolve()

        for root, exclude_spec in compiled_patterns.items():
            try:
                relpath = pathlib.Path(os.path.relpath(abspath, start=root))
            except ValueError:
                # We get here if path and exclude_root are on different drives (on
                # Windows).Things should still behave roughly as expected without
                # using a relative path. See
                # test_cli_utils.py::test_exclude_filter_windows for test cases.
                relpath = abspath

            if exclude_spec.match_file(relpath):
                return True
        return False

    return is_excluded


def resolve_sources(
    paths: List[pathlib.Path],
    exclude_patterns: List[ExcludePattern],
    extensions: List[str],
) -> List[Optional[pathlib.Path]]:
    """Resolves paths passed via CLI to a list of filepaths to lint.

    `paths` is a list of paths that may be files or directories. Files are
    returned verbatim if they exist, and directories are recursively searched
    for files that have an extension specified in `extensions`. Paths that match a
    pattern in `exclude_patterns` are ignored (based on gitignore pattern
    format, see https://git-scm.com/docs/gitignore#_pattern_format).

    Raises FileNotFoundError if a supplied path does not exist.
    """
    extensions = [f".{ext}" if not ext.startswith(".") else ext for ext in extensions]
    is_excluded = make_exclude_filter(exclude_patterns)

    sources: List[Optional[pathlib.Path]] = []

    for path in paths:
        if str(path) == "-":
            sources.append(None)
            continue

        if not path.exists():
            raise FileNotFoundError(f"path {path} does not exist")

        if is_excluded(path):
            continue

        if not path.is_dir():
            sources.append(path)
            continue

        for dirpath, _, filenames in os.walk(path):
            for name in filenames:
                _, ext = os.path.splitext(name)
                if ext.lower() in extensions:
                    child = pathlib.Path(dirpath) / name
                    if not is_excluded(child):
                        sources.append(child)

    return sources



from tclint.commands.checks import arg_count
from tclint.parser import Parser
from tclint import syntax_tree as ast


def test_arg_count():
    parser = Parser()
    assert arg_count([ast.BareWord("foo"), ast.BareWord("bar")], parser) == (2, False)
    assert arg_count([ast.ArgExpansion(ast.BracedWord("foo bar"))], parser) == (
        2,
        False,
    )
    # Returns minimum possible number of arguments.
    assert arg_count([ast.ArgExpansion(ast.VarSub("foo"))], parser) == (0, True)
